#!/usr/bin/env python
# -*- coding: utf-8 -*-

from covsirphy.__version__ import __version__

__citation__ = "Hirokazu Takaya and CovsirPhy Development Team (2020-2022), " \
    f"CovsirPhy version {__version__}: " \
    "Python library for COVID-19 analysis with phase-dependent SIR-derived ODE models, " \
    "https://github.com/lisphilar/covid19-sir"
